/*****************************************************************************/
/*!
\file      easyboot.c
\brief     intel hex bootloader for simple use with a terminal program 
                         

\implementation
  project       easy boot
  version       0.00
  copyright     cigarette lizense. Everybody can feel free in use and change
                of this code. If you think this little tool is fine, you can
                send a box of cigaretts to me. 
  
  generated     Matthias Koelling 26.03.2008
  review
  released

  compiler      AVR GCC
  target        ATmega128-16AU
\endimplementation

\history  
  Rev     Date         Author      Commentary
  0.00    26.03.2008   Koelling    generated
\endhistory  
*/
/*****************************************************************************/

/* -- Includes ------------------------------------------------------------- */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/boot.h>
#include <stdlib.h>
#include <util/delay.h>
// include additional MCU definitions to allow compiling
// for different MCUs from same source code
#include "mcudefs.h"

/* -- Configuration Parameters --------------------------------------------- */

/* -- Makros --------------------------------------------------------------- */

/* -- Defines -------------------------------------------------------------- */ 

// #define this to include code for programming the EEPROM
// it reduced bootloader code size by 124 bytes.
// code size with eeprom code:   1022 bytes
// code size without eeprom code: 898 bytes
#undef EEPROM_CODE
//#define EEPROM_CODE

// CPU frequency is normally defined in the Makefile
#ifndef F_CPU
   #define F_CPU 16000000
#endif

// TRUE and FALSE
#define TRUE  1
#define FALSE 0

// UART baudrate
//#define BAUDRATE 115200
#define BAUDRATE 115200
// memory type to program
#define NONE        (uint8_t)0
#define FLASH       (uint8_t)1
#define EEPROM      (uint8_t)2
#define NUM_REC_DU8 (uint8_t)5

#define MP3_ENABLE         PC3
#define MP3_ENABLE_DDR     DDC3

/* -- Types ---------------------------------------------------------------- */

/* -- Module Global Variables ---------------------------------------------- */

// ring buffer for received bytes, size of 256 eases the wrap around handling 
static uint8_t ReceiveBuffer_mau8[256];
// array with pointers to start of hex records in the ring buffer
static uint16_t ReadIndexArray_mau16[NUM_REC_DU8] = { 0, 0, 0, 0, 0 };

/* -- Extern Global Variables ---------------------------------------------- */

#ifdef EEPROM_CODE
   uint8_t eepromWriteBuffer[SPM_PAGESIZE+2];  // SPM_PAGESIZE used as eeprom buffer size
#endif
uint8_t srecEndOfFile = FALSE;
uint8_t programThisMemory = NONE;

/* -- Modul Global Function Prototypes ------------------------------------- */

void uartPutchar(uint8_t);  // put a uint8_tacter over USART
uint8_t uartGetchar(void);  // read (wait for) a uint8_tacter from USART
void uartPutHex(uint8_t);  // print hex byte over USART
void uartPutString(uint8_t *);
uint8_t hexuint8_tToInt(uint8_t);  // convert hex uint8_tacter to unsigned integer
uint8_t hexByteToInt(uint8_t, uint8_t);  // convert hex byte (two uint8_tacters) to unsigned integer
uint8_t ParseHexRecord(uint8_t index_u8);  // parse the current s-record in the buffer
void (*startApplication)( void ) = (void *)0x0000;  // pointer to application at flash start
//void (*startBootloader)( void ) = (void *)0xFC00;  // pointer to the bootloader start address
void (*startBootloader)( void ) = (void *)BOOTADR;  // pointer to the bootloader start address
/* -- Makros --------------------------------------------------------------- */

/* -- Implementation Functions --------------------------------------------- */

//-----------------------------------------------------------------------------
/*!
   \brief     ISR(myUSART_ReceiveCompleteVect)

      interrupt routine called when a uint8_tacter was received from USART
   
   \author     Koelling
   \date       26.03.2008

   \param[in]  void
   \param[out] none
   \return     void
*/
//-----------------------------------------------------------------------------

ISR(myUSART_ReceiveCompleteVect) 
{
   uint8_t c;
   static uint8_t array_index_hu8 = 0;
   static uint8_t byte_counter_hu8 = 0;
   static uint8_t byte_counter_temp_hu8 = 0;
   static uint8_t read_index_hu8 = 0;

   c = myUDR;  // save the uint8_tacter
   //            uartPutchar('.');  // '.' indicates some progress
         PORTC |= _BV(MP3_ENABLE);     

   if((c != '\r') && (c != '\n')) 
   {
      ReceiveBuffer_mau8[byte_counter_hu8] = c;  // write uint8_tacter to the buffer
      byte_counter_hu8++;
   } 
   else
   {
      // if there is '\r' and '\n' sent at the end of line, do the
      // following code only once
      if(byte_counter_temp_hu8 != byte_counter_hu8)
      {
         // mark that this job is already done
         byte_counter_temp_hu8 = byte_counter_hu8;
         // save the last beginning of a hex record in the buffer to the pointer array
         // this indicates a new valid record for the parser
         ReadIndexArray_mau16[array_index_hu8] = 0xFF00 + (uint16_t)read_index_hu8;
         // when last entry is reached restart at the beginning
         if(array_index_hu8 == NUM_REC_DU8)
         {
            array_index_hu8 = 0;
         }
         // save the current position in the receive buffer
         read_index_hu8 = byte_counter_hu8;
      }
   }

}


//-----------------------------------------------------------------------------
/*!
   \brief     int main(void)

      main routine. initialize hardware. start application
   
   \author     Koelling
   \date       26.03.2008

   \param[in]  void
   \param[out] none
   \return     void
*/
//-----------------------------------------------------------------------------

int main(void) 
{
   static uint8_t array_index_hu8 = 0;
   uint8_t c;
   uint16_t loop = 0;
   uint8_t bootloaderEnableFlag = FALSE;
   uint8_t array_index_u8 = 0;
   uint16_t index_u16 = 0;
   
   // relocate interrupt vector table to bottom of flash
   // in case the bootloader will not be started
   myIVSELREG = _BV(IVCE);
   myIVSELREG = 0;        

   // init USART
   myUBRRH = (F_CPU/(BAUDRATE*8L)-1) >> 8;             // calculate baudrate and set high byte
   myUBRRL = (uint8_t)(F_CPU/(BAUDRATE*8L)-1);         // and low byte
   myUCSRA |= myU2X;
   myUCSRB = _BV(myTXEN) | _BV(myRXEN) | _BV(myRXCIE);  // enable transmitter and receiver and receiver interrupt
   myUCSRC = myURSEL | _BV(myUCSZ1) | _BV(myUCSZ0);     // 8 bit character size, 1 stop bit, no parity
   DDRC |= _BV(MP3_ENABLE_DDR);

   // the bootloader may be activated either if
   // the uint8_tacter 'i' (interactive mode) was received from USART
   // or the flash is (still) empty

   // poll USART receive complete flag 64k times to catch the 'i' reliably
   do 
   {
      if(bit_is_set(myUCSRA, myRXC))
      {
         if(myUDR == 'i')
         {
            bootloaderEnableFlag = TRUE;
         }
      }
   } 
   while(--loop);
   
   // test if flash is empty (i.e. flash content is 0xff)
   if(pgm_read_byte_near(0x0000) == 0xFF) 
   {
      bootloaderEnableFlag = TRUE;  // set enable flag
   }

   // check enable flag and start application if FALSE
   if(bootloaderEnableFlag == FALSE) 
   {
      myUCSRB = 0;  // clear USART register to reset default
      startApplication();  // start application code
   }

   // now the bootloader code begins

   // welcome message and prompt
   uartPutchar('\r');
   uartPutchar('>');

   // loop until a valid uint8_tacter is received
   do 
   {
      c = uartGetchar();  // read a uint8_tacter

      if(c == 'f') // 'f' selects flash programming
      {  
         uartPutchar('f');           // echo the 'f'
         programThisMemory = FLASH;  // set flag
      }

      #ifdef EEPROM_CODE
         if(c == 'e') // 'e' selects eeprom programming
         {  
            uartPutchar('e');            // echo the 'e'
            programThisMemory = EEPROM;  // set flag
         }
      #endif

      if(c == 'g')              // 'g' starts the application
      {  
         uartPutchar('g');      // echo the 'g'
         startApplication();    // and jump to 0x0000
      }
   } 
   while(!programThisMemory);  // exit loop when a valid key was pressed


   // move interrupt vector table to boot loader area
   myIVSELREG = _BV(IVCE);
   myIVSELREG = _BV(IVSEL);
 
   //uartPutchar('\r');  // set cursor to next line

   // enable interrupts
   sei();

   // endless loop
   while(1)
   {
      index_u16 = ReadIndexArray_mau16[array_index_hu8];
  
      // if buffer is full, parse the buffer and write to flash
   
      if((index_u16 & 0xFF00) == 0xFF00) 
      {

         ReadIndexArray_mau16[array_index_hu8] = 0;

         //cli();  // disable interrupts

         // if parsing produced an error, restart bootloader
         if(!ParseHexRecord((uint8_t)index_u16)) 
         {
            cli();  // disable interrupts
            uartPutchar('\r');  // set cursor to next line
            startBootloader();  // restart the bootloader
         }

         // was an end-of-file s-record found?
         if(srecEndOfFile) 
         {
            cli();  // disable interrupts
           
            uartPutchar('O');  // 'OK' indicates successful programming
            uartPutchar('K');
            loop_until_bit_is_set(myUCSRA, myUDRE);  // wait until uint8_tacter is transmitted
            startBootloader();       // restart the bootloader
         }
         //sei();  // enable interrupts
      }
      array_index_hu8++;
      // when last entry is reached restart at the beginning
      if(array_index_hu8 == NUM_REC_DU8)
      {
         array_index_hu8 = 0;
      }
   }
}


//-----------------------------------------------------------------------------
/*!
   \brief     uint8_t ParseHexRecord(uint8_t ReceiveBuffer_mau8)

      parse the current hex-record
   
   \author     Koelling
   \date       27.03.2008

   \param[in]  void
   \param[out] none
   \return     void
*/
//-----------------------------------------------------------------------------

uint8_t ParseHexRecord(uint8_t index_u8) 
{
   static uint32_t linear_address_hu32 = 0;
   static uint32_t page_address_hu32 = 0;
   static uint8_t  first_record_hu8 = 1;
   static uint16_t last_word_hu16 = 0xFFFF;
   static uint8_t  byte_left_hu8 = 0;
   static myCOUNTERTYPE page_counter_ht = 0;
   myCOUNTERTYPE page_counter_t = 0;
   uint16_t tmp_word_u16;
   uint32_t linear_address_u32 = 0;
   uint8_t bytecount_u8;
   uint8_t checksum_u8;
   uint8_t record_type_u8;
   uint8_t tmp_address_u8;
   uint16_t record_address_u16;
   uint8_t high_byte_u8;
   uint8_t low_byte_u8;
   uint8_t c1, c2;
   uint16_t i = 0;
   uint8_t data_index_u8;
   
   // check if current buffer is a hex record (starts with :)
   uartPutchar(ReceiveBuffer_mau8[index_u8]);
   if(ReceiveBuffer_mau8[index_u8] == ':' ) 
   {
      // get record type
      high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 7];
      uartPutchar(high_byte_u8);
      low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 8];
      uartPutchar(low_byte_u8);
      uartPutchar('\n');
      record_type_u8 = hexByteToInt(high_byte_u8, low_byte_u8);
      checksum_u8 = record_type_u8;  // start checksum calculation
      
      // only process the record in case it's a data record
      switch(record_type_u8)
      {
      case 0:                    // data record
         // get the byte count
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 1];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 2];
         bytecount_u8 = hexByteToInt(high_byte_u8, low_byte_u8);
         checksum_u8 += bytecount_u8;  // add byte count to checksum
       
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 3];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 4];
         // get high byte of 16bit record address
         record_address_u16 = hexByteToInt(high_byte_u8, low_byte_u8); 
         // add the high byte to checksum
         checksum_u8 += (uint8_t)record_address_u16;
         record_address_u16 <<= 8;
         
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 5];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 6];
         // get low byte of 16bit record address
         tmp_address_u8 = hexByteToInt(high_byte_u8, low_byte_u8); 
         // add the low byte to checksum
         checksum_u8 += tmp_address_u8; 
         record_address_u16 += (uint16_t)tmp_address_u8;

         linear_address_u32 = linear_address_hu32 + record_address_u16;
         if((page_address_hu32 != (linear_address_u32 & (0xFFFFFFFF - (SPM_PAGESIZE - 1)))) &&
            // this ensures that the first record can start with an address other than 0
            (first_record_hu8 == 0)) 
         {
            boot_page_erase(page_address_hu32); // do a page erase
            boot_spm_busy_wait();              // wait for page erase done
            boot_page_write(page_address_hu32); // do a page write
            boot_spm_busy_wait();              // wait for write completed
            boot_rww_enable();                 // reenable rww section again
            page_address_hu32 = linear_address_u32 & (0xFFFFFFFF - (SPM_PAGESIZE - 1));    
         }
         else
         {
            // page counter counts words in the page  
            page_counter_t = (uint8_t)(linear_address_u32 & SPM_PAGESIZE);       
      
            data_index_u8 = index_u8 + 9; // set pointer to data field in the record
            // read all data bytes
            do
            {
               // if there was a lonly byte to fill in the page buffer with no
               // consecutive byte in the actual record
               if(byte_left_hu8 == 2)
               {
                  page_counter_t = (uint8_t)(linear_address_u32 & SPM_PAGESIZE);       
                  byte_left_hu8 = 0;
               }
               
               if(page_counter_t & 0x01) 
               {
                  // the address is odd, so we have to look if we must arrange the word with 
                  // the currently read byte and the temporary saved byte when the addresses
                  // are consecutive or if we must align a left byte to a word by setting the 
                  // high byte to 0xFF when the addresses are not consecutive 
                  if((page_counter_ht - page_counter_t) == 1)
                  {
                     // the addresses are consecutive 
                     // arrange the word with the left and the new byte
                     high_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                     low_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                     c1 = hexByteToInt(high_byte_u8, low_byte_u8);
                     checksum_u8 += c1;
                     i++;
                     tmp_word_u16 = last_word_hu16 | ((uint16_t)c1 << 8);
                     page_counter_t--; // correct page counter to even address
                  }
                  else
                  {
                     tmp_word_u16 = last_word_hu16 | 0xFF00;
                     page_counter_t = page_counter_ht; // take the old page_counter
                     byte_left_hu8 = 2; // mark that we have to reread the page_counter
                  }
               }
               else
               {
                  if(byte_left_hu8 == 1)
                  {
                     tmp_word_u16 = last_word_hu16 | 0xFF00;
                     page_counter_t = page_counter_ht; // take the old page_counter
                     byte_left_hu8 = 2; // mark that we have to reread the page_counter
                  }
                  else
                  {
                     // the address is even but we have to check whether there is another
                     // byte in the record to build a word to fill in the page
                     high_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                     low_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                     c1 = hexByteToInt(high_byte_u8, low_byte_u8);
                     checksum_u8 += c1;
                     i++;
                     if(i < bytecount_u8)
                     {
                        // there is a byte left, so we can  build a word
                        high_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                        low_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                        c2 = hexByteToInt(high_byte_u8, low_byte_u8);
                        checksum_u8 += c2;
                        tmp_word_u16 = ((uint16_t)c2 << 8) + (uint16_t)c1;  
                        i++;
                     }
                     else
                     {
                        // there is no other byte, so we have to save the lonly byte till 
                        // processing the next record
                        last_word_hu16 = (uint16_t)c1;
                        page_counter_ht = page_counter_t; // save current page counter
                        // look in the next record if there is a consecutive byte
                        byte_left_hu8 = 1;                
                        // get checksum and compare to 0xff
                        high_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                        low_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
                        checksum_u8 += hexByteToInt(high_byte_u8, low_byte_u8);  
                        // compare checksum to 0xff
                        if(checksum_u8 != 0) 
                        {
                           uartPutchar('C');  // if checksum is wrong, 'C' indicates this error
                           return FALSE;
                        }
                        return FALSE; // not MISRA conform but easier to handle
                     }
                  }
               }
               #ifdef EEPROM_CODE
                  if(programThisMemory == FLASH) 
                  {
               #endif
                  // write word to flash write buffer
                  boot_page_fill(page_counter_t, tmp_word_u16);
               #ifdef EEPROM_CODE
                  } 
                  else 
                  {
                     // write hi and low_byte_u8 byte into eeprom buffer
                     eepromWriteBuffer[page_counter_t] = (uint8_t)tmp_word_u16;
                     eepromWriteBuffer[page_counter_t + 1] = (uint8_t)(tmp_word_u16 >> 8);
                  }
               #endif

               page_counter_t += 2;
               if(page_counter_t == SPM_PAGESIZE)
               {
                  boot_page_erase(page_address_hu32); // do a page erase
                  boot_spm_busy_wait();              // wait for page erase done
                  boot_page_write(page_address_hu32); // do a page write
                  boot_spm_busy_wait();              // wait for write completed
                  boot_rww_enable();                 // reenable rww section again
                  page_counter_t = 0;
               }
            }
            while(i < bytecount_u8);
            
            // get checksum and compare to 0xff
            high_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
            low_byte_u8 = ReceiveBuffer_mau8[data_index_u8++];
            checksum_u8 += hexByteToInt(high_byte_u8, low_byte_u8);  
            // compare checksum to 0xff
            if(checksum_u8 != 0) 
            {
               uartPutchar('C');  // if checksum is wrong, 'C' indicates this error
               return FALSE;
            }
            uartPutchar('.');  // '.' indicates some progress
         }
         first_record_hu8 = 1; // must be placed after else
         break;
      case 1:  // end of file record
         srecEndOfFile = TRUE;
         break;
      case 4:                    // extended linear address record
         // get the byte count
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 1];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 2];
         bytecount_u8 = hexByteToInt(high_byte_u8, low_byte_u8);
         
         if(bytecount_u8 != 4)
         {
            uartPutchar('L');  // if length is wrong, 'L' indicates this error
            return FALSE;
         }
         checksum_u8 += bytecount_u8;  // add byte count to checksum
         
         
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 9];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 10];
         // get high byte of 16bit record address
         record_address_u16 = hexByteToInt(high_byte_u8, low_byte_u8); 
         // add the high byte to checksum
         checksum_u8 += (uint8_t)record_address_u16;
         record_address_u16 <<= 8;
         
         high_byte_u8 = ReceiveBuffer_mau8[index_u8 + 11];
         low_byte_u8 = ReceiveBuffer_mau8[index_u8 + 12];
         // get low byte of 16bit record address
         tmp_address_u8 = hexByteToInt(high_byte_u8, low_byte_u8); 
         // add the low byte to checksum
         checksum_u8 += tmp_address_u8; 
         record_address_u16 += (uint16_t)tmp_address_u8;
         
         linear_address_hu32 =  (uint32_t)record_address_u16 << 16;
      
         // compare checksum to 0xff
         if(checksum_u8 != 0xff) 
         {
            uartPutchar('C');  // if checksum is wrong, 'C' indicates this error
            return FALSE;
         }
         break;
      default:
         return FALSE;
      } // end switch(record_type_u8)
   } // end if(ReceiveBuffer_mau8 == ':' )
   return TRUE;
}


//-----------------------------------------------------------------------------
/*!
   \brief     uint8_t hexuint8_tToInt(uint8_t c)

      convert 1 ASCII coded byte into unsigned uint8_tacter (low nibble)
   
   \author     
   \date       

   \param[in]  uint8_t c -> ASCII coded uint8_tacter
   \param[out] none
   \return     uint8_t   -> uint8_tacter in 4 bit binary format
*/
//-----------------------------------------------------------------------------

uint8_t hexuint8_tToInt(uint8_t c) 
{
   // test if uint8_tacter is a letter or a number
   if(c <= '9')
   {
      return (c - '0');
   }
   else
   {
      return (c - 'A' + 0xa);
   }
}


//-----------------------------------------------------------------------------
/*!
   \brief     uint8_t hexByteToInt(uint8_t high_byte_u8, uint8_t low_byte_u8)

      convert 2 ASCII coded bytes into 8 bit unsigned uint8_tacter
   
   \author     
   \date       

   \param[in]  uint8_t high_byte_u8 -> ASCII coded high byte
               uint8_t low_byte_u8  -> ASCII coded low byte
   \param[out] none
   \return     uint8_t              -> unsigned uint8_tacter
*/
//-----------------------------------------------------------------------------

uint8_t hexByteToInt(uint8_t high_byte_u8, uint8_t low_byte_u8) 
{
  return ((hexuint8_tToInt(high_byte_u8) << 4) + hexuint8_tToInt(low_byte_u8));  
}


//-----------------------------------------------------------------------------
/*!
   \brief     void uartPutchar(uint8_t c)

      send a uint8_tacter through the USART
   
   \author     
   \date       

   \param[in]  uint8_t c -> uint8_tacter to send
   \param[out] none
   \return     void
*/
//-----------------------------------------------------------------------------

void uartPutchar(uint8_t c) 
{
   if(c == '\r')                               // if uint8_tacter is 'caraige return', then
   {
      uartPutchar('\n');                       // send an additional 'new line'
   }
   loop_until_bit_is_set(myUCSRA, myUDRE);  // wait until transmit buffer is empty
   myUDR = c;                               // write uint8_tacter to transmit buffer
}


//-----------------------------------------------------------------------------
/*!
   \brief     uint8_t uartGetchar(void)

      receive a uint8_tacter from the USART
   
   \author     
   \date       

   \param[in]  void
   \param[out] none
   \return     uint8_t -> received uint8_tacter
*/
//-----------------------------------------------------------------------------

uint8_t uartGetchar(void) 
{
   loop_until_bit_is_set(myUCSRA, myRXC);  // wait until uint8_tacter is received
   return myUDR;                           // return the uint8_tacter
}

// eof
